/* Raw image loader (and saver) plugin 3.0
 *
 * by tim copperfield [timecop@japan.co.jp]
 * http://www.ne.jp/asahi/linux/timecop
 *
 * This plugin is not based on any other plugin.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <libgimp/gimpintl.h>

typedef enum {
    RAW_RGB,			/* RGB Image */
    RAW_RGBA,			/* RGB Image with an Alpha channel */
    RAW_PLANAR,			/* Planar RGB */
    RAW_INDEXED			/* Indexed image */
} raw_type;

typedef enum {
    RAW_PALETTE_RGB,		/* standard RGB */
    RAW_PALETTE_BGR		/* Windows BGRX */
} raw_palette_type;

typedef struct {
    gint32 file_offset;		/* offset to beginning of image in raw data */
    gint32 image_width;		/* width of the raw image */
    gint32 image_height;	/* height of the raw image */
    raw_type image_type;	/* type of image (RGB, INDEXED, etc) */
    gint32 palette_offset;	/* offset inside the palette file, if any */
    raw_palette_type palette_type;	/* type of palette (RGB/BGR) */
} raw_config;

typedef struct {
    FILE *fp;			/* pointer to the already open file */
    GimpDrawable *drawable;	/* gimp drawable */
    GimpPixelRgn region;	/* gimp pixel region */
    gint32 image_id;		/* gimp image id */
    guchar cmap[768];		/* color map for indexed images */
} raw_gimp_data;

raw_config *runtime;		/* main config struct */
int file;			/* descriptor for the preview image */
GtkWidget *preview;		/* preview widget so I dont have to pass it all the fucking time */
char *palfile = NULL;		/* filename for the indexed palette */
gboolean result = FALSE;	/* this is set to true after OK click in any
				   dialog */

/* prototypes for the new load functions */
gboolean raw_load_standard(raw_gimp_data * data, gint bpp);
gboolean raw_load_planar(raw_gimp_data * data);
gboolean raw_load_palette(raw_gimp_data * data, gchar * palette_filename);

/* support functions */
static gint32 get_file_info(gchar * filename);
static void raw_read_row(FILE * fp, guchar * buf, gint32 offset, gint32 size);
static int mmap_read(int fd, void *buf, gint32 len, gint32 pos, gint rowstride);
static gint32 load_image(gchar * filename);
static gint32 save_image(gchar * filename, gint32 image_id, gint32 drawable_id);

/* gui functions */
static void callback_ok(GtkWidget * widget, gpointer data);
static void palette_ok(GtkWidget * widget, GtkWidget ** fs);
static void my_int_update(GtkAdjustment * adj, gpointer data);
static void my_radio_update(GtkWidget * widget, gpointer data);
static void filesel_callback(GtkWidget * widget, gpointer data);
void preview_update(void);
static gint load_dialog(gchar * name);
static gint save_dialog(gchar * filename, gint32 image_id, gint32 drawable_id);
static void palette_dialog(gchar * title);

/* gimp function prototypes */
static void query(void);
static void run(gchar * name,
		gint nparams,
		GimpParam * param,
		gint * nreturn_vals, GimpParam ** return_vals);

/* end prototypes */

GimpPlugInInfo PLUG_IN_INFO = {
    NULL,			/* init_proc  */
    NULL,			/* quit_proc  */
    query,			/* query_proc */
    run,			/* run_proc   */
};

MAIN()

/* get file size from a filename */
gint32 get_file_info(gchar * filename)
{
    struct stat *status;
    gint32 size;

    status = g_malloc(sizeof(struct stat));
    stat(filename, status);
    size = status->st_size;
    g_free(status);
    return size;
}
    
/* new image handle functions */
static void raw_read_row(FILE * fp, guchar * buf, gint32 offset,
			 gint32 size)
{
    fseek(fp, offset, SEEK_SET);
    if (!fread(buf, size, 1, fp)) {
	fprintf(stderr, "fread failed\n");
	memset(buf, 0xFF, size);
    }
}

/* similar to the above function but memset is done differently. has nothing
 * to do with mmap, by the way */
static int mmap_read(int fd, void *buf, gint32 len, gint32 pos,
		     gint rowstride)
{
    lseek(fd, pos, SEEK_SET);
    if (!read(fd, buf, len))
	memset(buf, 0xFF, rowstride);
    return 0;
}

gboolean raw_load_standard(raw_gimp_data * data, gint bpp)
    /* this handles 1, 3, 4 bpp "standard" images */
{
    guchar *row = NULL;

    row = g_malloc(runtime->image_width * runtime->image_height * bpp);
    if (!row)
	return FALSE;

    raw_read_row(data->fp, row, runtime->file_offset, (runtime->image_width * runtime->image_height * bpp));
    gimp_pixel_rgn_set_rect(&data->region, row, 0, 0, runtime->image_width, runtime->image_height);
    g_free(row);
    return TRUE;
}

gboolean raw_load_planar(raw_gimp_data * data)
    /* this handles 3 bpp "planar" images */
{
    gint32 r_offset, g_offset, b_offset, i, j, k;
    guchar *r_row, *b_row, *g_row, *row;
    gint bpp = 3; /* adding support for alpha channel should be easy */

    /* red, green, blue rows temporary data */
    r_row = g_malloc(runtime->image_width);
    g_row = g_malloc(runtime->image_width);
    b_row = g_malloc(runtime->image_width);

    /* row for the pixel region, after combining RGB together */
    row = g_malloc(runtime->image_width * bpp);

    r_offset = runtime->file_offset;
    g_offset = r_offset + runtime->image_width * runtime->image_height;
    b_offset = g_offset + runtime->image_width * runtime->image_height;

    for (i = 0; i < runtime->image_height; i++) {
	/* Read R, G, B rows */
	raw_read_row(data->fp, r_row,
		     r_offset + (runtime->image_width * i),
		     runtime->image_width);
	raw_read_row(data->fp, g_row,
		     g_offset + (runtime->image_width * i),
		     runtime->image_width);
	raw_read_row(data->fp, b_row,
		     b_offset + (runtime->image_width * i),
		     runtime->image_width);
	/* Combine separate R, G and B rows into RGB triples */
	for (j = 0, k = 0; j < runtime->image_width; j++) {
	    row[k++] = r_row[j];
	    row[k++] = g_row[j];
	    row[k++] = b_row[j];
	}
	gimp_pixel_rgn_set_row(&data->region, row, 0, i,
			       runtime->image_width);
	gimp_progress_update((gfloat) i / (gfloat) runtime->image_height);
    }
    g_free(row);
    g_free(r_row);
    g_free(g_row);
    g_free(b_row);
    return TRUE;
}

gboolean raw_load_palette(raw_gimp_data * data, gchar * palette_file)
{
    guchar temp[1024];
    int fd, i, j;

    if (palette_file) {
	fd = open(palette_file, O_RDONLY);
	if (!fd)
	    return FALSE;
	lseek(fd, runtime->palette_offset, SEEK_SET);

	switch (runtime->palette_type) {
	case RAW_PALETTE_RGB:
	    read(fd, data->cmap, 768);
	    break;
	case RAW_PALETTE_BGR:
	    read(fd, temp, 1024);
	    for (i = 0, j = 0; i < 256; i++) {
		data->cmap[j++] = temp[i * 4 + 2];
		data->cmap[j++] = temp[i * 4 + 1];
		data->cmap[j++] = temp[i * 4];
	    }
	    break;
	}
	close(fd);
    } else {
	/* make a fake grayscale color map */
	for (i = 0, j = 0; i < 256; i++) {
	    data->cmap[j++] = i;
	    data->cmap[j++] = i;
	    data->cmap[j++] = i;
	}
    }
    gimp_image_set_cmap(data->image_id, data->cmap, 256);
    return TRUE;
}

/* end new image handle functions */

static void query(void)
{
    static GimpParamDef load_args[] = {
	{GIMP_PDB_INT32, "run_mode", "Interactive"},
	{GIMP_PDB_STRING, "filename", "The name of the file to load"},
	{GIMP_PDB_STRING, "raw_filename", "The name entered"},
    };
    static gint nload_args = sizeof(load_args) / sizeof(load_args[0]);

    static GimpParamDef load_return_vals[] = {
	{GIMP_PDB_IMAGE, "image", "Output image"}
    };
    static gint nload_return_vals = (sizeof(load_return_vals) /
				     sizeof(load_return_vals[0]));

    static GimpParamDef save_args[] = {
	{GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
	{GIMP_PDB_IMAGE, "image", "Input image"},
	{GIMP_PDB_DRAWABLE, "drawable", "Drawable to save"},
	
	    {GIMP_PDB_STRING, "filename",
	 "The name of the file to save the image in"},
	{GIMP_PDB_STRING, "raw_filename", "The name entered"},
    };

    static gint nsave_args = sizeof(save_args) / sizeof(save_args[0]);

    gimp_install_procedure("file_raw_load",
			   "Load raw images, specifying image information",
			   "Load raw images, specifying image information",
			   "timecop",
			   "timecop",
			   "Oct 2000",
			   N_("<Load>/RAW"),
			   NULL,
			   GIMP_PLUGIN,
			   nload_args, nload_return_vals,
			   load_args, load_return_vals);

    gimp_install_procedure("file_raw_save",
			   "Dump images to disk in raw format",
			   "Dump images to disk in raw format",
			   "timecop",
			   "timecop",
			   "Oct 2000",
			   N_("<Save>/RAW"),
			   "INDEXED, GRAY, RGB, RGBA",
			   GIMP_PLUGIN, nsave_args, 0, save_args, NULL);

    gimp_register_magic_load_handler("file_raw_load", "raw", "", "");
    gimp_register_save_handler("file_raw_save", "raw", "");
}

static void
run(gchar * name,
    gint nparams,
    GimpParam * param, gint * nreturn_vals, GimpParam ** return_vals)
{

    static GimpParam values[2];
    GimpRunModeType run_mode;
    GimpPDBStatusType status = GIMP_PDB_SUCCESS;
    gint32 image_id;
    gint32 drawable_id;

    run_mode = param[0].data.d_int32;

    *nreturn_vals = 1;
    *return_vals = values;
    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;

    /* allocate config structure and fill with defaults */
    runtime = g_malloc0(sizeof(raw_config));
    runtime->file_offset = 0;
    runtime->image_width = 128;
    runtime->image_height = 128;
    runtime->palette_offset = 0;
    runtime->palette_type = RAW_PALETTE_RGB;

    /* load raw image */
    if (strcmp(name, "file_raw_load") == 0) {
	if (run_mode == GIMP_RUN_INTERACTIVE) {
	    gimp_get_data("file_raw_load", runtime);
	    if (!load_dialog(param[1].data.d_string)) {
		status = GIMP_PDB_CANCEL;
	    }
	} else {
	    /* we only run interactively due to the nature of this plugin.
	     * things like generate preview etc like to call us non-
	     * interactively.  here we stop that. */
	    status = GIMP_PDB_CALLING_ERROR;
	}
	/* we are okay, and the user clicked OK in the load dialog */
	if (status == GIMP_PDB_SUCCESS) {
	    image_id = load_image(param[1].data.d_string);
	    if (image_id != -1) {
		gimp_set_data("file_raw_load", runtime, sizeof(raw_config));
		*nreturn_vals = 2;
		values[1].type = GIMP_PDB_IMAGE;
		values[1].data.d_image = image_id;
	    } else {
		status = GIMP_PDB_CALLING_ERROR;
	    }
	}
    /* save raw image */
    } else if (strcmp(name, "file_raw_save") == 0) {
	image_id = param[1].data.d_int32;
	drawable_id = param[2].data.d_int32;

	if (run_mode == GIMP_RUN_INTERACTIVE) {
	    gimp_get_data("file_raw_save", runtime);
	    if (nparams != 5)
		status = GIMP_PDB_CALLING_ERROR;
	    if (!save_dialog
		(param[3].data.d_string, image_id, drawable_id)) {
		status = GIMP_PDB_CANCEL;
	    }
	} else {
	    /* we want to make sure we always do save-as for raw images
	     * to avoid confusion and data loss */
	    status = GIMP_PDB_CALLING_ERROR;
	}

	if (status == GIMP_PDB_SUCCESS) {
	    status = save_image(param[3].data.d_string, image_id, drawable_id);
	}
    }
    values[0].data.d_status = status;
    g_free(runtime);
}

gint32 save_image(gchar * filename, gint32 image_id, gint32 drawable_id)
{
    GimpDrawable *drawable;
    GimpPixelRgn pixel_rgn;
    guchar *cmap = NULL;	/* colormap for indexed images */
    guchar *buf;
    guchar *red, *green, *blue, *alpha = NULL;
    gint32 width, height, bpp = 0, have_alpha = 0;
    FILE *fp;
    int i, j = 0;
    gint palsize = 0;
    gint32 ret = GIMP_PDB_EXECUTION_ERROR;	/* return status */

    /* get info about the current image */
    drawable = gimp_drawable_get(drawable_id);
    switch (gimp_drawable_type(drawable_id)) {
    case GIMP_RGB_IMAGE:
	bpp = 3;
	break;
    case GIMP_RGBA_IMAGE:
	bpp = 4;
	have_alpha = 1;
	break;
    case GIMP_GRAY_IMAGE:
	bpp = 1;
	break;
    case GIMP_GRAYA_IMAGE:
	bpp = 1;
	have_alpha = 1;
	break;
    case GIMP_INDEXED_IMAGE:
	bpp = 1;
	cmap = gimp_image_get_cmap(image_id, &palsize);
	break;
    case GIMP_INDEXEDA_IMAGE:
	bpp = 1;
	have_alpha = 1;
	cmap = gimp_image_get_cmap(image_id, &palsize);
	break;
    }
    width = drawable->width;
    height = drawable->height;

    gimp_pixel_rgn_init(&pixel_rgn, drawable, 0, 0, drawable->width,
			drawable->height, FALSE, FALSE);

    buf = (guchar *) g_malloc(width * height * bpp);
    gimp_pixel_rgn_get_rect(&pixel_rgn, buf, 0, 0, width, height);
    fp = fopen(filename, "wb");
    if (!fp)
	return GIMP_PDB_EXECUTION_ERROR;

    ret = GIMP_PDB_SUCCESS;

    switch (runtime->image_type) {
    case RAW_RGB:
	if (!fwrite(buf, width * height * bpp, 1, fp)) {
	    return GIMP_PDB_EXECUTION_ERROR;
	}
	fclose(fp);
	if (cmap) {		/* we have a colormap, too.write it into filename+pal */
	    char *newfile;
	    newfile = g_malloc0(strlen(filename) + 5);
	    strncpy(newfile, filename, strlen(filename));
	    strncat(newfile, ".pal", 4);
	    fp = fopen(newfile, "wb");
	    if (!fp)
		return GIMP_PDB_EXECUTION_ERROR;
	    switch (runtime->palette_type) {
		char *temp;
	    case RAW_PALETTE_RGB:
		if (!fwrite(cmap, palsize * 3, 1, fp))
		    ret = GIMP_PDB_EXECUTION_ERROR;
		fclose(fp);
		break;
	    case RAW_PALETTE_BGR:
		temp = g_malloc0(palsize * 4);
		for (i = 0, j = 0; i < palsize * 3; i += 3) {
		    temp[j++] = cmap[i + 2];
		    temp[j++] = cmap[i + 1];
		    temp[j++] = cmap[i];
		    temp[j++] = 0;
		}
		if (!fwrite(temp, palsize * 4, 1, fp))
		    ret = GIMP_PDB_EXECUTION_ERROR;
		fclose(fp);
		g_free(temp);
		break;
	    }
	}
	break;
    case RAW_PLANAR:
	red = (guchar *) g_malloc(width * height);
	green = (guchar *) g_malloc(width * height);
	blue = (guchar *) g_malloc(width * height);
	if (have_alpha)
	    alpha = (guchar *) g_malloc(width * height);
	for (i = 0; i < width * height * bpp; i += bpp) {
	    red[j] = buf[i];
	    green[j] = buf[i + 1];
	    blue[j] = buf[i + 2];
	    if (have_alpha)
		alpha[j] = buf[i + 3];
	    j++;
	}
	ret = GIMP_PDB_SUCCESS;
	if (!fwrite(red, width * height, 1, fp))
	    ret = GIMP_PDB_EXECUTION_ERROR;
	if (!fwrite(green, width * height, 1, fp))
	    ret = GIMP_PDB_EXECUTION_ERROR;
	if (!fwrite(blue, width * height, 1, fp))
	    ret = GIMP_PDB_EXECUTION_ERROR;
	if (have_alpha) {
	    if (!fwrite(alpha, width * height, 1, fp))
		ret = GIMP_PDB_EXECUTION_ERROR;
	}
	g_free(red);
	g_free(green);
	g_free(blue);
	if (have_alpha)
	    g_free(alpha);
	fclose(fp);
	break;
    default:
	/* we can never get here but we gotta do this cause image_type is a
	 * enum */
	break;
    }
    return ret;
}

gint32 load_image(gchar * filename)
{
    /* main struct with the image info in it */
    raw_gimp_data *data;
    /* stuff for the gimp image, local to this function */
    gint32 layer_id = -1;
    GimpImageType ltype = GIMP_RGB;
    GimpImageBaseType itype = GIMP_RGB_IMAGE;

    gint32 size;
    gint bpp = 0;

    gchar *name_buf;

    /* allocate the raw_gimp_data structure */
    data = g_malloc0(sizeof(raw_gimp_data));

    /* gimp fuckup. exit */
    data->fp = fopen(filename, "rb");
    if (!data->fp) {
	g_message(_("RAW: Can't open \"%s\""), filename);
	return -1;
    }

    name_buf = g_strdup_printf(_("Loading %s:"), filename);
    gimp_progress_init(name_buf);
    g_free(name_buf);

    size = get_file_info(filename);

    switch (runtime->image_type) {
    case RAW_RGB:		/* standard RGB */
    case RAW_PLANAR:		/* planar RGB */
	bpp = 3;
	ltype = GIMP_RGB_IMAGE;
	itype = GIMP_RGB;
	break;
    case RAW_RGBA:		/* RGB + alpha */
	bpp = 4;
	ltype = GIMP_RGBA_IMAGE;
	itype = GIMP_RGB;
	break;
    case RAW_INDEXED:		/* Indexed */
	bpp = 1;
	ltype = GIMP_INDEXED_IMAGE;
	itype = GIMP_INDEXED;
	break;
    }
    /* make sure we don't load image bigger than file size */
    if (runtime->image_height > (size / runtime->image_width / bpp))
	runtime->image_height = size / runtime->image_width / bpp;

    /* create new gimp image */
    data->image_id =
	gimp_image_new(runtime->image_width, runtime->image_height, itype);
    /* set image filename */
    gimp_image_set_filename(data->image_id, filename);
    /* create new layer - local variable */
    layer_id =
	gimp_layer_new(data->image_id, _("Background"),
		       runtime->image_width, runtime->image_height, ltype,
		       100, GIMP_NORMAL_MODE);
    gimp_image_add_layer(data->image_id, layer_id, 0);
    /* create gimp drawable for the image */
    data->drawable = gimp_drawable_get(layer_id);
    /* initialize pixel region */
    gimp_pixel_rgn_init(&data->region, data->drawable, 0, 0,
			runtime->image_width, runtime->image_height, TRUE,
			FALSE);

    switch (runtime->image_type) {
    case RAW_RGB:
    case RAW_RGBA:
	raw_load_standard(data, bpp);
	break;
    case RAW_PLANAR:
	raw_load_planar(data);
	break;
    case RAW_INDEXED:
	raw_load_palette(data, palfile);
	raw_load_standard(data, bpp);
	break;
    }
    fclose(data->fp);

    gimp_drawable_flush(data->drawable);
    gimp_drawable_detach(data->drawable);

    return data->image_id;
}

/* misc GUI stuff */
static void palette_ok(GtkWidget * widget, GtkWidget ** fs)
{
    if (palfile)
	g_free(palfile);

    palfile =
	g_strdup(gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)));
    preview_update();
    gtk_widget_destroy(GTK_WIDGET(fs));
}

static void callback_ok(GtkWidget * widget, gpointer data)
{
    result = TRUE;
    gtk_widget_destroy(GTK_WIDGET(data));
}

static void my_int_update(GtkAdjustment * adj, gpointer data)
{
    gimp_int_adjustment_update(adj, data);
    preview_update();
}

static void my_radio_update(GtkWidget * widget, gpointer data)
{
    gimp_radio_button_update(widget, data);
    preview_update();
}

static void filesel_callback(GtkWidget * widget, gpointer data)
{
    palette_dialog(_("Select pallete file to load"));
}

void preview_update(void)
{
    gint psize;
    gint width;
    gint height;
    gint32 pos, len;
    gint x, y;

    /* buffers for various images */
    guchar *even;
    guchar *odd;
    guchar *r_row, *g_row, *b_row;
    /* for alpha and indexed images */
    guchar *alpha;
    guchar *index;
    /* for doing RGBA loading, mainly */
    guchar *p0, *p1;
    gdouble c0;
    gdouble c1;
    gdouble pixel[4];

    /* colormap */
    static guchar cmap[1024];
    static gint32 oldpal;
    static raw_palette_type oldptype;

    if (!preview)
	return;

    psize = (GTK_PREVIEW(preview)->buffer_width) *
	(GTK_PREVIEW(preview)->buffer_height) *
	(GTK_PREVIEW(preview)->bpp);

    width = (runtime->image_width > 256) ? 256 : runtime->image_width;
    height = (runtime->image_height > 256) ? 256 : runtime->image_height;
    memset(GTK_PREVIEW(preview)->buffer, 0xFF, psize);	/* white BG */

    even = g_malloc(256 * 3);
    odd = g_malloc(256 * 3);


    switch (runtime->image_type) {
    case RAW_RGB:
	/* standard RGB image */
	len = width * 3;
	for (y = 0; y < height; y++) {
	    pos = runtime->file_offset + runtime->image_width * y * 3;
	    mmap_read(file, even, len, pos, width * 3);
	    gtk_preview_draw_row(GTK_PREVIEW(preview), even, 0, y, width);
	}
	gtk_widget_queue_draw(preview);
	break;
    case RAW_PLANAR:
	r_row = g_malloc(256);
	g_row = g_malloc(256);
	b_row = g_malloc(256);
	for (y = 0; y < height; y++) {
	    int j, k;
	    memset(r_row, 0, 256);
	    memset(even, 0, 256 * 3);
	    pos = runtime->file_offset + (y * runtime->image_width);
	    mmap_read(file, r_row, width, pos, 256);
	    pos =
		runtime->file_offset +
		(runtime->image_width * (runtime->image_height + y));
	    mmap_read(file, g_row, width, pos, 256);
	    pos =
		runtime->file_offset +
		(runtime->image_width * (runtime->image_height * 2 + y));
	    mmap_read(file, b_row, width, pos, 256);
	    for (j = 0, k = 0; j < width; j++) {
		even[k++] = r_row[j];
		even[k++] = g_row[j];
		even[k++] = b_row[j];
	    }
	    gtk_preview_draw_row(GTK_PREVIEW(preview), even, 0, y, width);
	}
	gtk_widget_queue_draw(preview);
	g_free(b_row);
	g_free(g_row);
	g_free(r_row);
	break;
    case RAW_RGBA:
	/* RGB + alpha image */
	len = width * 4;
	alpha = g_malloc(256 * 4);
	for (y = 0; y < height; y++) {
	    int draw = (y / GIMP_CHECK_SIZE) & 1;
	    p0 = even;
	    p1 = odd;
	    pos = runtime->file_offset + runtime->image_width * y * 4;
	    mmap_read(file, alpha, len, pos, width * 4);
	    for (x = 0; x < width; x++) {
		pixel[0] = ((gdouble) alpha[x * 4]) / 255.0;
		pixel[1] = ((gdouble) alpha[x * 4 + 1]) / 255.0;
		pixel[2] = ((gdouble) alpha[x * 4 + 2]) / 255.0;
		pixel[3] = ((gdouble) alpha[x * 4 + 3]) / 255.0;

		if ((x / GIMP_CHECK_SIZE) & 1) {
		    c0 = GIMP_CHECK_LIGHT;
		    c1 = GIMP_CHECK_DARK;
		} else {
		    c0 = GIMP_CHECK_DARK;
		    c1 = GIMP_CHECK_LIGHT;
		}

		/* we draw either even or odd.  So we only calculate
		 * the ones we need and skip the others forward */
		if (draw) {
		    *p1++ = (c1 + (pixel[0] - c1) * pixel[3]) * 255.0;
		    *p1++ = (c1 + (pixel[1] - c1) * pixel[3]) * 255.0;
		    *p1++ = (c1 + (pixel[2] - c1) * pixel[3]) * 255.0;
		    p0 += 3;
		} else {
		    *p0++ = (c0 + (pixel[0] - c0) * pixel[3]) * 255.0;
		    *p0++ = (c0 + (pixel[1] - c0) * pixel[3]) * 255.0;
		    *p0++ = (c0 + (pixel[2] - c0) * pixel[3]) * 255.0;
		    p1 += 3;
		}
	    }
	    if (draw)
		gtk_preview_draw_row(GTK_PREVIEW(preview), odd, 0, y,
				     width);
	    else
		gtk_preview_draw_row(GTK_PREVIEW(preview), even, 0, y,
				     width);
	}
	gtk_widget_queue_draw(preview);
	g_free(alpha);
	break;
    case RAW_INDEXED:
	/* indexed image */
	len = width;
	index = g_malloc(256);
	if ( ((oldptype != runtime->palette_type) || (oldpal != runtime->palette_offset)) && palfile ) {
	    int fd;
	    fd = open(palfile, O_RDONLY);
	    lseek(fd, runtime->palette_offset, SEEK_SET);
	    read(fd, cmap,
		 (runtime->palette_type == RAW_PALETTE_RGB) ? 768 : 1024);
	    close(fd);
	    oldpal = runtime->palette_offset;
	    oldptype = runtime->palette_type;
	}
	if (!palfile) {
	    /* make fake palette, maybe overwrite it later */
	    for (y = 0, x = 0; y < 256; y++) {
		cmap[x++] = y;
		cmap[x++] = y;
		if (runtime->palette_type == RAW_PALETTE_RGB) {
		    cmap[x++] = y;
		} else {
		    cmap[x++] = y;
		    cmap[x++] = 0;
		}
	    }
	}

	for (y = 0; y < height; y++) {
	    p0 = even;
	    pos = runtime->file_offset + runtime->image_width * y;
	    mmap_read(file, index, len, pos, width);
	    for (x = 0; x < width; x++) {
		switch (runtime->palette_type) {
		case RAW_PALETTE_RGB:
		    *p0++ = cmap[index[x] * 3];
		    *p0++ = cmap[index[x] * 3 + 1];
		    *p0++ = cmap[index[x] * 3 + 2];
		    break;
		case RAW_PALETTE_BGR:
		    *p0++ = cmap[index[x] * 4 + 2];
		    *p0++ = cmap[index[x] * 4 + 1];
		    *p0++ = cmap[index[x] * 4];
		    break;
		}
	    }
	    gtk_preview_draw_row(GTK_PREVIEW(preview), even, 0, y, width);
	}
	gtk_widget_queue_draw(preview);
	g_free(index);
	break;
    }
    g_free(even);
    g_free(odd);
}

gint load_dialog(gchar * name)
{
    GtkWidget *dialog;
    GtkWidget *table;
    GtkWidget *frame;
    GtkWidget *abox;
    GtkWidget *label;
    GtkWidget *spinbutton;
    GtkObject *adj;
    GtkWidget *filesel;

    gint32 size;


    size = get_file_info(name);
    file = open(name, O_RDONLY);

    if (!file) {
	fprintf(stderr, "<rant>\n"
		"Once upon a time someone decided that if you\n"
		"choose a file type in Gimp Load Image dialog\n"
		"then it would \"automagically\" append the extension\n"
		"to the file name, replacing the current, and correct\n"
		"extension with whatever the plugin was supposed to\n"
		"handle.  So what happens, is instead of a perfectly\n"
		"existing filename, we receive GARBAGE.  Garbage in,\n"
		"Garbage Out.\n"
		"You can safely ignore the next message about messing\n"
		"up Gimp.  It's already messed up :)\n" "</rant>\n");
	perror("blame gimp");
	exit(0);
    }

    gimp_ui_init("raw", TRUE);

    dialog = gimp_dialog_new(_("Raw Image Loader 3.0"), "raw",
			     gimp_standard_help_func, "filters/raw.html",
			     GTK_WIN_POS_MOUSE,
			     FALSE, TRUE, FALSE,
			     _("OK"), callback_ok, NULL, NULL, NULL, TRUE,
			     FALSE, _("Cancel"), gtk_widget_destroy, NULL,
			     1, NULL, FALSE, TRUE, NULL);

    gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
		       GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

    table = gtk_table_new(8, 2, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 6);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    gtk_table_set_col_spacings(GTK_TABLE(table), 4);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), table,
		       FALSE, FALSE, 0);
    gtk_widget_show(table);

    frame = gtk_frame_new(_("Preview"));
    gtk_container_set_border_width(GTK_CONTAINER(frame), 0);
    gtk_widget_show(frame);
    gtk_table_attach(GTK_TABLE(table), frame, 0, 2, 0, 1, GTK_FILL,
		     GTK_FILL, 0, 0);

    abox = gtk_alignment_new(0.5, 0.5, 0.0, 0.0);
    gtk_container_set_border_width(GTK_CONTAINER(abox), 4);
    gtk_container_add(GTK_CONTAINER(frame), abox);
    gtk_widget_show(abox);
    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
    gtk_container_add(GTK_CONTAINER(abox), frame);
    gtk_widget_show(frame);
    /* preview is global */
    preview = gtk_preview_new(GTK_PREVIEW_COLOR);

    gtk_preview_size(GTK_PREVIEW(preview), 256, 256);
    gtk_container_add(GTK_CONTAINER(frame), preview);
    gtk_widget_show(preview);
    {
	/* fill the preview with crap. Seems to be needed */
	int row;
	guchar gay[256 * 3];
	memset(gay, 0xFF, 256 * 3);
	for (row = 0; row < 256; row++)
	    gtk_preview_draw_row(GTK_PREVIEW(preview), gay, 0, row, 256);
    }
    preview_update();

    frame = gimp_radio_group_new2(TRUE, _("Image type"),
				  my_radio_update,
				  &runtime->image_type,
				  (gpointer) runtime->image_type,
				  _("RGB"), RAW_RGB, NULL,
				  _("RGBA"), RAW_RGBA, NULL,
				  _("Planar RGB"), RAW_PLANAR, NULL,
				  _("Indexed"), RAW_INDEXED, NULL, NULL);
    gtk_table_attach(GTK_TABLE(table), frame, 0, 1, 1, 5, GTK_FILL,
		     GTK_FILL, 0, 0);
    gtk_widget_show(frame);

    label = gtk_label_new(_("Offset:"));
    spinbutton =
	gimp_spin_button_new(&adj, runtime->file_offset, 0, size, 1, 1000,
			     0, 1, 0);
    gtk_table_attach(GTK_TABLE(table), label, 1, 2, 1, 2, GTK_FILL,
		     GTK_FILL, 0, 0);
    gtk_table_attach(GTK_TABLE(table), spinbutton, 1, 2, 2, 3, GTK_FILL,
		     GTK_FILL, 0, 0);
    gtk_signal_connect(GTK_OBJECT(adj), "value_changed",
		       GTK_SIGNAL_FUNC(my_int_update),
		       &runtime->file_offset);
    gtk_widget_show(spinbutton);
    gtk_widget_show(label);

    label = gtk_label_new(_("Width:"));
    spinbutton =
	gimp_spin_button_new(&adj, runtime->image_width, 1, 4096, 1, 10, 0,
			     1, 0);
    gtk_table_attach(GTK_TABLE(table), label, 1, 2, 3, 4, GTK_FILL,
		     GTK_FILL, 0, 0);
    gtk_table_attach(GTK_TABLE(table), spinbutton, 1, 2, 4, 5, GTK_FILL,
		     GTK_FILL, 0, 0);
    gtk_signal_connect(GTK_OBJECT(adj), "value_changed",
		       GTK_SIGNAL_FUNC(my_int_update),
		       &runtime->image_width);
    gtk_widget_show(spinbutton);
    gtk_widget_show(label);

    label = gtk_label_new(_("Height:"));
    spinbutton =
	gimp_spin_button_new(&adj, runtime->image_height, 0, 4096, 1, 10,
			     0, 1, 0);
    gtk_table_attach(GTK_TABLE(table), label, 1, 2, 5, 6, GTK_FILL,
		     GTK_FILL, 0, 0);
    gtk_table_attach(GTK_TABLE(table), spinbutton, 1, 2, 6, 7, GTK_FILL,
		     GTK_FILL, 0, 0);
    gtk_signal_connect(GTK_OBJECT(adj), "value_changed",
		       GTK_SIGNAL_FUNC(my_int_update),
		       &runtime->image_height);
    gtk_widget_show(spinbutton);
    gtk_widget_show(label);

    label = gtk_label_new(_("Palette Offset:"));
    spinbutton =
	gimp_spin_button_new(&adj, runtime->palette_offset, 0, 1 << 24, 1,
			     768, 0, 1, 0);
    gtk_table_attach(GTK_TABLE(table), label, 1, 2, 7, 8, GTK_FILL,
		     GTK_FILL, 0, 0);
    gtk_table_attach(GTK_TABLE(table), spinbutton, 1, 2, 8, 9, GTK_FILL,
		     GTK_FILL, 0, 0);
    gtk_signal_connect(GTK_OBJECT(adj), "value_changed",
		       GTK_SIGNAL_FUNC(my_int_update),
		       &runtime->palette_offset);
    gtk_widget_show(spinbutton);
    gtk_widget_show(label);

    frame = gimp_radio_group_new2(TRUE, _("Pallete Type"),
				  my_radio_update,
				  &runtime->palette_type,
				  (gpointer) runtime->palette_type,
				  _("R, G, B (normal)"), RAW_PALETTE_RGB,
				  NULL, _("B, G, R, X (bmp style)"),
				  RAW_PALETTE_BGR, NULL, NULL);
    gtk_table_attach(GTK_TABLE(table), frame, 0, 1, 7, 9, GTK_FILL,
		     GTK_FILL, 0, 0);
    gtk_widget_show(frame);

    filesel = gtk_button_new_with_label(_("Load Palette..."));
    gtk_signal_connect(GTK_OBJECT(filesel), "clicked",
		       GTK_SIGNAL_FUNC(filesel_callback), NULL);
    gtk_table_attach(GTK_TABLE(table), filesel, 0, 1, 6, 7, GTK_FILL,
		     GTK_FILL, 0, 0);
    gtk_widget_show(filesel);

    gtk_widget_show(dialog);

    gtk_main();
    gdk_flush();

    return result;
}

static gint save_dialog(gchar * filename, gint32 image_id,
			gint32 drawable_id)
{
    GtkWidget *dialog;
    GtkWidget *frame;

    gimp_ui_init("raw", TRUE);

    dialog = gimp_dialog_new(_("Raw Image Save"), "raw",
			     gimp_standard_help_func, "filters/raw.html",
			     GTK_WIN_POS_MOUSE,
			     FALSE, TRUE, FALSE,
			     _("OK"), callback_ok, NULL, NULL, NULL, TRUE,
			     FALSE, _("Cancel"), gtk_widget_destroy, NULL,
			     1, NULL, FALSE, TRUE, NULL);

    gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
		       GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

    frame = gimp_radio_group_new2(TRUE, _("RGB Save Type"),
				  gimp_radio_button_update,
				  &runtime->image_type,
				  (gpointer) runtime->image_type,
				  _("Standard (R,G,B)"), RAW_RGB,
				  NULL, _("Planar (RRR,GGG,BBB)"),
				  RAW_PLANAR, NULL, NULL);
    gtk_container_set_border_width(GTK_CONTAINER(frame), 4);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), frame,
		       FALSE, FALSE, 0);

    gtk_widget_show(frame);

    frame = gimp_radio_group_new2(TRUE, _("Indexed Pallete Type"),
				  gimp_radio_button_update,
				  &runtime->palette_type,
				  (gpointer) runtime->palette_type,
				  _("R, G, B (normal)"), RAW_PALETTE_RGB, NULL,
				  _("B, G, R, X (bmp style)"), RAW_PALETTE_BGR, NULL,
				  NULL);
    gtk_container_set_border_width(GTK_CONTAINER(frame), 4);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), frame,
		       FALSE, FALSE, 0);

    gtk_widget_show(frame);
    gtk_widget_show(dialog);
    gtk_main();
    gdk_flush();
    return result;
}

static void palette_dialog(gchar * title)
{
    GtkWidget *dialog;

    dialog = gtk_file_selection_new(title);
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(dialog), "");

    gtk_signal_connect
	(GTK_OBJECT(GTK_FILE_SELECTION(dialog)->ok_button), "clicked",
	 GTK_SIGNAL_FUNC(palette_ok), dialog);
    gtk_signal_connect_object
	(GTK_OBJECT(GTK_FILE_SELECTION(dialog)->cancel_button), "clicked",
	 GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));

    gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
		       GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

    gtk_widget_show(dialog);

    gtk_main();
    gdk_flush();
}
